"""
Import this module to monkey patch various classes for display in an IPython
HTML notebook.
"""
from collections import OrderedDict, Counter, defaultdict
import asyncio
from asyncio.selector_events import BaseSelectorEventLoop
from brittle_wit import TwitterRequest


def _make_table(d, header_labels=None, sort=False):
    snippets = ["<table>"]

    if header_labels:
        assert len(header_labels) == 2
        snippets.append("<tr><th>{}</th><th>{}</th>".format(*header_labels))

    row_str = "<tr><td>{}</td><td>{}</td>"

    ks = sorted(d) if sort else d
    for k in ks:
        snippets.append(row_str.format(k, d[k]))

    snippets.append("</table>")

    return "\n".join(snippets)


def _make_ol(items):
    if not items:
        return ""

    return "<ul>\n\t<li>" + "</li>\n\t<li>".join(items) + "\t</li>\n</ul>"


def twitter_request_html_repr(self):
    d = OrderedDict()

    for k in ['_method', '_url', '_family', '_service', '_parse_as']:
        d[k[1:]] = getattr(self, k)

    if self._params:
        d['Params'] = _make_table(self._params, ('Key', 'Value'), True)

    api_url = TwitterRequest.DOC_URLS.get(self._url, None)
    if api_url:
        api_link = "<a href='{}' target='_new'>{}</a>".format(api_url, api_url)
        d["docs"] = api_link.replace(".json", "")

    return _make_table(d, ("Name", "Value"))


def event_loop_html_repr(self):
    """
    Note: This only works if you run_forever() otherwise tasks get cleared.
    """
    d = OrderedDict()

    d['Type'] = type(self).__name__
    d['Is Running'] = self.is_running()
    d['Is Closed'] = self.is_closed()
    d['Is Debug'] = self.get_debug()
    d['Time'] = self.time()
    d['Current Task'] = asyncio.Task.current_task(self)

    tasks = asyncio.Task.all_tasks(self)
    pending, n_done, errored, coro_to_err = [], 0, [], defaultdict(set)

    for t in tasks:
        name = t._coro.cr_code.co_name
        if t.done():
            err = t.exception()
            if err:
                errored.append((name, err))
                coro_to_err[name].add(repr(err))
            n_done += 1
        else:
            pending.append(name)

    d['N(Tasks)'] = len(tasks)
    d['N(Done)'] = n_done
    d['N(Pending)'] = len(pending)
    if pending:
        d['Pending'] = _make_table(Counter(pending), ("Name", "N"),
                                  sort=True)

    d['N(Err)'] = len(errored)
    if errored:
        d['Errors'] = _make_table(coro_to_err,
                                 ("Coroutine", "Exceptions"),
                                 sort=True)

    return _make_table(d, ('Name', 'Value'))


BaseSelectorEventLoop._repr_html_ = event_loop_html_repr
TwitterRequest._repr_html_ = twitter_request_html_repr



"""
This module contains a bare-bones OAuth implementation. It is
minimally-compliant so as to conform to Twitter's requirements.

See: https://dev.twitter.com/oauth/overview
"""

import binascii
import hashlib
import hmac
import random
import time

from functools import total_ordering
from urllib.parse import quote as urllib_quote

from brittle_wit import __version__

ANY_CREDENTIALS = None

ALPHANUMERIC = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"
USER_AGENT = 'BrittleWit/' + __version__


def _quote(s):
    if type(s) is int or type(s) is float:
        s = str(s)
    elif s is True:
        s = 'true'
    elif s is False:
        s = 'false'

    return urllib_quote(s, safe='')


def _generate_nonce(length=42):
    """
    Generate an alpha numeric string that is unique for each request.

    Twitter used a 42 character alpha-numeric (case-sensitive) string in the
    API documentation. However, they note "any approach which produces a
    relatively random alphanumeric string should be OK here." I opted not to
    use a cryptographically secure source of entropy. `SystemRandom` is
    convenient, but it uses file IO to connect to `/dev/urandom`. Adding
    `async` machinery here seems like expensive complexity.
    """
    return "".join(random.choice(ALPHANUMERIC) for _ in range(length))


def _generate_timestamp():
    return int(time.time())


def _generate_header_string(params):
    """
    Generate the string for use in the http "Authorization" header field.

    :param params: a dictionary with the oauth_* key-values.
    """
    return "OAuth " + ", ".join('{}="{}"'.format(k, _quote(params[k]))
                                for k in sorted(params))


def _generate_param_string(params):
    """
    Generate the parameter string for signing.

    :param params: A dictionary with both the request and oauth_* parameters,
        less the `oauth_signature`.
    """
    d = {_quote(k): _quote(v) for k, v in params.items()}
    return "&".join(["{}={}".format(k, d[k]) for k in sorted(d)])


def _generate_sig_base_string(method, base_url, param_string):
    """
    :param method: either `get` or `post`
    :param base_url: the API base URL (i.e. without parameters)
    :param param_string: string generated by generate_param_string
    """
    return "&".join([method.upper(), _quote(base_url), _quote(param_string)])


def _generate_signing_key(consumer_secret, oauth_token_secret):
    return "{}&{}".format(_quote(consumer_secret), _quote(oauth_token_secret))


def _generate_signature(sig_base_string, signing_key):
    digest = hmac.new(signing_key.encode('ascii'),
                      sig_base_string.encode('ascii'),
                      hashlib.sha1).digest()

    return binascii.b2a_base64(digest)[:-1]  # Strip newline


def generate_req_headers(twitter_req, app_cred, client_cred, **overrides):
    """
    Generate the 'Authorization' and 'User-Agent' headers.

    :param twitter_req: A TwitterRequest object
    :param app_cred: an AppCredentials object
    :param client_cred: a ClientCredentials object
    :param overrides: key-value pairs which override (or, adds a new value
        to) the oauth_* dictionary used for signature generation.
    """
    oauth_d = {'oauth_consumer_key': app_cred.key,
               'oauth_nonce': _generate_nonce(),
               'oauth_signature_method': "HMAC-SHA1",
               'oauth_timestamp': str(_generate_timestamp()),
               'oauth_token': client_cred.token,
               'oauth_version': "1.0"}

    if overrides:
        oauth_d.update(overrides)

    param_string = _generate_param_string({**oauth_d, **twitter_req.params})
    sig_base_string = _generate_sig_base_string(twitter_req.method,
                                                twitter_req.url,
                                                param_string)
    signing_key = _generate_signing_key(app_cred.secret, client_cred.secret)
    oauth_d['oauth_signature'] = _generate_signature(sig_base_string,
                                                     signing_key)

    return {'Authorization': _generate_header_string(oauth_d),
            'User-Agent': USER_AGENT}


class AppCredentials:
    """
    An Immutable set of application credentials.
    """

    __slots__ = '_key', '_secret'

    def __init__(self, key, secret):
        self._key, self._secret = key, secret

    @property
    def key(self):
        return self._key

    @property
    def secret(self):
        return self._secret

    def __str__(self):
        s = "AppCredentials({}, {})"
        return s.format(self._key, "*" * len(self._secret))

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash((self._key, self._secret))

    def __eq__(self, other):
        return self.key == self.key and self.secret == other.secret

    def __ne__(self, other):
        return not self == other


@total_ordering
class ClientCredentials:
    """
    An Immutable set of client credentials.

    Note: Equality testing and hashing is a function of the user_id alone!
    """

    __slots__ = '_user_id', '_token', '_secret'

    def __init__(self, user_id, token, secret):
        self._user_id, self._token, self._secret = user_id, token, secret

    @property
    def token(self):
        return self._token

    @property
    def secret(self):
        return self._secret

    @property
    def user_id(self):
        return self._user_id

    @property
    def as_dict(self):
        return {'user_id': self._user_id,
                'token': self._token,
                'secret': self._secret}

    @staticmethod
    def from_dict(d):
        return ClientCredentials(d['user_id'], d['token'], d['secret'])

    def __str__(self):
        s = "ClientCredentials({}, {}, {})"
        return s.format(self.user_id, self._token, "*" * len(self._secret))

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self._user_id)

    def __eq__(self, other):
        return self.user_id == other.user_id

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        # Because the client_credentials may partially-determine ordering
        # in a priority queue.
        return self.user_id < other.user_id



import json
import unittest
from functools import lru_cache

from brittle_wit import *
from brittle_wit import TwitterRequest
from brittle_wit.oauth import (_generate_nonce,
                               _generate_timestamp,
                               _generate_header_string,
                               _generate_param_string,
                               _generate_sig_base_string,
                               _generate_signing_key,
                               _generate_signature,
                               _quote,
                               generate_req_headers,
                               AppCredentials)


class TestOAuth(unittest.TestCase):
    """
    Test the OAuth functions against the test data and expectations provided
    by Twitter's API documentation.

    See: https://dev.twitter.com/oauth/overview
    """
    def test_quote(self):
        self.assertEqual(_quote(1), "1")
        self.assertEqual(_quote(1.0), "1.0")
        self.assertEqual(_quote(True), "true")
        self.assertEqual(_quote(False), "false")

    def test_generate_nonce(self):
        self.assertEqual(len(_generate_nonce(100)), 100)
        self.assertNotEqual(_generate_nonce(), _generate_nonce())

    def test_generate_timestamp(self):
        self.assertTrue(type(_generate_timestamp()) == int)

    def test_generate_header_string(self):
        params = _load_oauth_params()
        expected = load_fixture_expectation("header_string.txt")
        self.assertEqual(_generate_header_string(params), expected)

    def test_generate_param_string(self):
        params = _load_request_params()
        expected = load_fixture_expectation("param_string.txt")
        self.assertEqual(_generate_param_string(params), expected)

    def test_generate_sig_base_string(self):
        method = "post"  # Keep lowercase as test of uppercase assurance
        url = "https://api.twitter.com/1/statuses/update.json"
        param_string = load_fixture_expectation("param_string.txt")

        result = _generate_sig_base_string(method, url, param_string)
        expected = load_fixture_expectation("sig_base_string.txt")

        self.assertEqual(result, expected)

    def test_generate_signing_key(self):
        consumer_secret = "kAcSOqF21Fu85e7zjz7ZN2U4ZRhfV3WpwPAoE3Z7kBw"
        token_secret = "LswwdoUaIvS8ltyTt5jkRh4J50vUPVVHtR2YPi5kE"
        k = _generate_signing_key(consumer_secret, token_secret)
        expected = load_fixture_expectation("signing_key.txt")
        self.assertEqual(k, expected)

    def test_generate_signature(self):
        signing_key = load_fixture_expectation("signing_key.txt")
        sig_base_string = load_fixture_expectation("sig_base_string.txt")
        expected = b"tnnArxj06cWHq44gCs1OSKk/jLY="

        self.assertEqual(_generate_signature(sig_base_string, signing_key),
                         expected)

    def test_generate_req_headers(self):
        oauth_params = _load_oauth_params()

        app = AppCredentials(oauth_params['oauth_consumer_key'],
                             "kAcSOqF21Fu85e7zjz7ZN2U4ZRhfV3WpwPAoE3Z7kBw")
        client = ClientCredentials(1,
                                   oauth_params['oauth_token'],
                                   "LswwdoUaIvS8ltyTt5jkRh4J50vUPVVHtR2YPi5kE")

        status = "Hello Ladies + Gentlemen, a signed OAuth request!"
        req = TwitterRequest("POST",
                             "https://api.twitter.com/1/statuses/update.json",
                             'statuses',
                             'statuses/update',
                             include_entities='true',
                             status=status)
        expected = load_fixture_expectation("header_string.txt")

        overrides = {k: oauth_params[k]
                     for k in ['oauth_nonce', 'oauth_timestamp']}

        auth = generate_req_headers(req, app, client, **overrides)
        self.assertIn('Authorization', auth.keys())
        self.assertEqual(auth['Authorization'], expected)


class TestAppCredentials(unittest.TestCase):
    def test_app_credentials(self):
        app_1 = AppCredentials("app_1", "secret")
        self.assertEqual(app_1, AppCredentials("app_1", "secret"))

        app_2 = AppCredentials("app_2", "password")
        self.assertNotEqual(app_1, app_2)

        self.assertEqual(len({app_1, app_2}), 2)
        self.assertEqual(str(app_1), "AppCredentials(app_1, ******)")
        self.assertEqual(repr(app_1), "AppCredentials(app_1, ******)")

        with self.assertRaises(AttributeError):
            app_1.key = 10  # Immutable(ish)


class TestClientCredentials(unittest.TestCase):
    def test_client_credentials(self):
        client_1 = ClientCredentials(1, "token_1", "secret")
        self.assertEqual(client_1, ClientCredentials(1, "token_1", "secret"))

        client_2 = ClientCredentials(2, "token_2", "secret")
        self.assertNotEqual(client_1, client_2)

        self.assertEqual(len({client_1, client_2}), 2)
        self.assertEqual(str(client_1),
                         "ClientCredentials(1, token_1, ******)")
        self.assertEqual(repr(client_1),
                         "ClientCredentials(1, token_1, ******)")

        with self.assertRaises(AttributeError):
            client_1.token = 10  # Immutable(ish)

        self.assertTrue(client_2 > client_1)

    def test_dict_serialization(self):
        client_1 = ClientCredentials(1, "token_1", "secret")
        client_2 = ClientCredentials.from_dict(client_1.as_dict)

        self.assertEqual(client_1.user_id, client_2.user_id)
        self.assertEqual(client_1.secret, client_2.secret)
        self.assertEqual(client_1.token, client_2.token)


FIXTURES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            "fixtures")

@lru_cache()
def _load_oauth_params():
    with open(os.path.join(FIXTURES_DIR, "oauth_params.json")) as fp:
        return json.load(fp)


@lru_cache()
def _load_request_params():
    with open(os.path.join(FIXTURES_DIR, "request_params.json")) as fp:
        return json.load(fp)


@lru_cache()
def load_fixture_expectation(file_name):
    print(file_name)
    with open(os.path.join(FIXTURES_DIR, file_name)) as fp:
        return fp.read()


if __name__ == '__main__':
    unittest.main()



import unittest
from brittle_wit.ticketing import TicketMaster
from test.helpers import *


async def use_ticket(ticket, identifier, busy_work):
    async with ticket:
        await busy_work
        return identifier


class BlowUpError(Exception):
    pass


async def blow_up():
    raise BlowUpError()


async def sim_ticket_master_usage(loop, n):
    tm = TicketMaster()

    tasks = []
    for i in range(n):
        # The first ticketed coro waits the longest....
        # The second waits second longest...
        busy_work = asyncio.sleep((n-1) * 0.02, loop=loop)

        coro = use_ticket(tm.take_ticket('test', loop), i, busy_work)
        tasks.append(loop.create_task(coro))
    results = []

    for task in asyncio.as_completed(tasks, loop=loop):
        results.append(await task)

    return results


async def sim_ticket_master_usage_with_blowup(loop):
    n = 5
    tm = TicketMaster()

    tasks = []
    for i in range(n):
        # The first ticketed coro waits the longest....
        # The second waits second longest..
        if i == 3:
            busy_work = blow_up()
        else:
            busy_work = asyncio.sleep((n-1) * 0.02, loop=loop)

        coro = use_ticket(tm.take_ticket('test', loop), i, busy_work)
        tasks.append(loop.create_task(coro))
    results = []

    for task in asyncio.as_completed(tasks, loop=loop):
        try:
            results.append(await task)
        except BlowUpError:
            pass

    return results


class TestTicketMaster(unittest.TestCase):
    def test_ticket_master(self):
        loop = asyncio.new_event_loop()
        n = 5
        result = loop.run_until_complete(sim_ticket_master_usage(loop, n))
        loop.close()
        self.assertEqual(result, list(range(n)))

    def test_ticket_master_with_failing_coro(self):
        loop = asyncio.new_event_loop()
        coro = sim_ticket_master_usage_with_blowup(loop)
        result = loop.run_until_complete(coro)
        loop.close()

        # Number 3 blows up
        self.assertEqual(result, [0, 1, 2, 4])

    def test_is_line_empty(self):
        tm = TicketMaster()
        self.assertTrue(tm.is_line_empty('main_line'))
        self.assertTrue(tm.is_line_empty('second_line'))
        tm._lines['main_line'] = 'MOCK'
        self.assertFalse(tm.is_line_empty('main_line'))
        self.assertTrue(tm.is_line_empty('second_line'))

    def test_all_lines_empty(self):
        tm = TicketMaster()
        self.assertTrue(tm.all_lines_empty())
        tm._lines['main_line'] = 'MOCK'
        self.assertFalse(tm.all_lines_empty())



from datetime import datetime
from functools import wraps


def parse_date(s):
    return datetime.strptime(s, "%a %b %d %H:%M:%S %z %Y")


class WrappingHandler:
    def __init__(self, underlying_func):
        self._underlying_func = underlying_func

    def send(self, msg):
        return self._underlying_func(msg)



import unittest
from unittest.mock import MagicMock

import aiohttp

from brittle_wit import (TwitterRequest, Cursor, TwitterResponse,
                         BrittleWitError)
from brittle_wit.messages import TwitterError, WrappedException
from brittle_wit.oauth import ClientCredentials


class TestBrittleWitError(unittest.TestCase):
    def test_brittle_wit_error(self):
        self.assertFalse(BrittleWitError().is_retryable)


class TestTwitterError(unittest.TestCase):

    def test_rate_limit_error(self):
        client_cred = ClientCredentials(1, "token", "secret")
        twitter_req = TwitterRequest('POST', "REQ-URL", "FAM", "SVC")
        resp = MagicMock()
        resp.status = 429

        err = TwitterError(client_cred, twitter_req, resp, "m")
        self.assertEqual(str(err), "TwitterError(code=429, msg=m)")
        self.assertEqual(str(err), repr(err))
        self.assertFalse(err.is_retryable)
        self.assertIs(err.twitter_req, twitter_req)
        self.assertIs(err.credentials, client_cred)

    def test_internal_server_error(self):
        client_cred = ClientCredentials(1, "token", "secret")
        twitter_req = TwitterRequest('POST', "REQ-URL", "FAM", "SVC")
        resp = MagicMock()
        resp.status = 500

        err = TwitterError(client_cred, twitter_req, resp, "m")
        self.assertEqual(str(err), "TwitterError(code=500, msg=m)")
        self.assertEqual(str(err), repr(err))
        self.assertTrue(err.is_retryable)
        self.assertEqual(err.message, "m")


class TestWrappedException(unittest.TestCase):
    def test_client_os_error(self):
        e = aiohttp.errors.ClientOSError()
        err = WrappedException(e)
        self.assertTrue(err.is_retryable)
        self.assertIs(err.underlying_exception, e)

    def test_keyboard_interrupt(self):
        err = WrappedException(KeyboardInterrupt())
        self.assertFalse(err.is_retryable)
        self.assertEqual(repr(err), "WrappedException(KeyboardInterrupt())")


class TestTwitterRequest(unittest.TestCase):
    def test_twitter_request(self):
        req = TwitterRequest("GET", "http://twitter.com/faux/service",
                             "faux", "faux/service", parse_as="json",
                             timeout=1000)
        req.parse_as = "raw"
        self.assertEqual(req.parse_as, "raw")
        self.assertEqual(str(req),
                         "TwitterRequest(url=http://twitter.com/faux/service)")

        dup = req.clone_and_merge({'timeout': 0})
        self.assertEqual(dup.params["timeout"], 0)
        self.assertEqual(req.params["timeout"], 1000)


class TestCursor(unittest.TestCase):
    def test_cursor(self):
        req = TwitterRequest("GET", "http://twitter.com/f/s", "f", "f/s")

        cursor = iter(Cursor(req))
        self.assertIs(req, next(cursor))
        with self.assertRaises(RuntimeError):
            next(cursor)

        cursor.update(TwitterResponse(None, None, {'next_cursor': 0}))
        with self.assertRaises(StopIteration):
            next(cursor)

        cursor = iter(Cursor(req))
        cursor.update(TwitterResponse(None, None, {}))
        with self.assertRaises(StopIteration):
            next(cursor)

        cursor = iter(Cursor(req))
        cursor.update(TwitterResponse(None, None, {'next_cursor': 100}))
        req = next(cursor)
        self.assertEqual(req.params['cursor'], 100)



import asyncio
import aiohttp
import time

from brittle_wit.constants import FIFTEEN_MINUTES, LOGGER
from brittle_wit.oauth import generate_req_headers, ANY_CREDENTIALS
from brittle_wit.rate_limit import RateLimit
from brittle_wit.ticketing import TicketMaster
from brittle_wit.messages import (TwitterResponse, TwitterError,
                                  WrappedException, BrittleWitError)


def twitter_req_to_http_req(session, app_cred, client_cred, twitter_req):
    headers = generate_req_headers(twitter_req, app_cred, client_cred)
    
    if twitter_req.method == 'POST':
        return session.request(twitter_req.method,
                               twitter_req.url,
                               data=twitter_req.params,
                               headers=headers)
    else:
        return session.request(twitter_req.method,
                               twitter_req.url,
                               params=twitter_req.params,
                               headers=headers)


async def execute_req(credentials, twitter_req, http_req, rate_limit, timeout):
    """
    Execute a request with no exception handling.

    :param credentials: the credentials used to build the http_req
    :param twitter_req: the original TwitterRequest
    :param http_req: the HttpRequest with encoded authentication information
    :param rate_limit: the RateLimit object associated with the endpoint and
        credentials
    :param timeout: the time, in seconds, to wait before an exception is
        thrown
    :return: a TwitterResponse
    """
    LOGGER.info("Executing %s request (%s) for %s",
                credentials.user_id, id(twitter_req), twitter_req.service)
    try:
        with aiohttp.Timeout(timeout):

            # Send request over the wire.
            async with http_req as resp:

                # Update rate limits ASAP. Note: The prior call to comply
                # decremented the rate limits already. If the context manager
                # threw an exception on entry, it is still in compliance.
                rate_limit.update(resp)

                # An error occurred.
                if resp.status != 200:
                    raise TwitterError(credentials,
                                       twitter_req,
                                       resp,
                                       await resp.json())

                # The status code indicates a good response.
                if twitter_req.parse_as == 'json':
                    body = await resp.json()
                else:
                    body = await resp.text()

                return TwitterResponse(twitter_req, resp, body)
    except TwitterError as e:
        LOGGER.exception("Failed %s request (%s) for %s with %s",
                         credentials.user_id,
                         twitter_req.service,
                         id(twitter_req),
                         e.status_code)
        raise e
    except Exception as e:
        LOGGER.exception("Failed %s request (%s) for %s",
                         credentials.user_id,
                         twitter_req.service,
                         id(twitter_req))
        raise WrappedException(e)


class ClientRequestProcessor:
    """
    The ClientRequestProcessor manages rate limits while executing requests.

    When the caller issues a request for execution, the request processor
    enforces rate limitations. Such enforcement occurs in the context of the
    given credentials and for a given service. If multiple callers request
    execution at the same time (for a given set of credentials and service),
    the requests execute FIFO.
    """

    def __init__(self, client_credentials):
        """
        :param client_credentials: the ClientCredentials to manage
        """
        self._client_credentials = client_credentials
        self._ticket_master = TicketMaster()
        self._rate_limits = {}

    @property
    def user_id(self):
        return self._client_credentials.user_id

    def _rate_limit_for(self, service):
        """
        Get existing or create new RateLimit for the given service.

        :param service: API endpoint service
        :return: the RateLimit object associated with the service
        """
        limit = self._rate_limits.get(service)

        if limit is not None:
            return limit
        else:
            max_requests = RateLimit.CLIENT_LIMITS[service]
            limit = RateLimit(max_requests,
                              max_requests,
                              time.time() + FIFTEEN_MINUTES)
            self._rate_limits[service] = limit
            return limit

    def cleanup(self, now=None):
        """
        Remove all rate limits which have expired.

        :param now: the current time as per time.time()
        """
        rate_limits = self._rate_limits

        # Abort quickly if possible.
        if not rate_limits:
            return

        now = now or time.time()

        remove_ks = []
        for service, rate_limit in rate_limits.items():
            # If the reset time is in the past, remove it.
            if rate_limit.reset_time < now:
                remove_ks.append(service)

        for k in remove_ks:
            del rate_limits[k]

    def is_removable(self):
        """
        :return: True if this executor is not in use and is monitoring no
            rate limits
        """
        return self.is_not_processing() and not self._rate_limits

    def is_not_processing(self):
        """
        :return: True if no caller is using this processors' credentials for
            any endpoint otherwise True
        """
        return self._ticket_master.all_lines_empty()

    def is_immediately_available_for(self, service):
        """
        :param service: the API service name, as per the Twitter docs
        :return: True if no caller is using this objects associated credentials
            for the given service AND that service is not rate limited
        """
        if self._ticket_master.is_line_empty(service):
            if service not in self._rate_limits:
                return True
            else:
                return not self._rate_limits[service].is_exhausted
        return False

    async def _execute(self, session, app_cred, twitter_req, timeout=120):
        service = twitter_req.service

        LOGGER.debug("Waiting for ticket on %s request (%s) for %s",
                     self._client_credentials.user_id,
                     id(twitter_req),
                     twitter_req.service)
        async with self._ticket_master.take_ticket(service):
            # The caller is now in charge of this service.
            rate_limit = self._rate_limit_for(service)

            # Block request execution until not rate limited.
            await rate_limit.comply()

            http_req = twitter_req_to_http_req(session,
                                               app_cred,
                                               self._client_credentials,
                                               twitter_req)

            # This coroutine will raise an exception for non-200 statuses.
            return await execute_req(self._client_credentials,
                                     twitter_req,
                                     http_req,
                                     rate_limit,
                                     timeout)

    async def execute(self, session, app_cred, twitter_req, timeout=120,
                      n_retries=5):
        tries = 0
        while True:
            try:
                return await self._execute(session, app_cred,
                                           twitter_req, timeout)
            except BrittleWitError as e:
                tries += 1
                if not e.is_retryable or tries >= n_retries:
                    raise e
                else:
                    await asyncio.sleep(60 * tries)  # Sleep progressively


class ContextualizedProcessor:
    def __init__(self, session, app_cred, client_processor):
        self._session = session
        self._app_cred = app_cred
        self._client_processor = client_processor

    @property
    def user_id(self):
        return self._client_processor.user_id

    def __call__(self, twitter_req, timeout=60):
        return self._client_processor.execute(self._session,
                                              self._app_cred,
                                              twitter_req,
                                              timeout)


class ManagedClientRequestProcessors:
    """

    = Any Credential Management

    For many tasks, you don't need specific credentials, you just need some
    credentials. For example, if you and your social network analyst friends
    wanted to analyze the followers of Bernie Sanders, you need to collect all
    the user objects. Given that there are millions of them that could take
    time. But, if you already have the user id's you can partition the
    collection and run it in parallel. The specific credentials don't matter.
    (Well, ignoring the case of a follower blocking that person.)

    Finding an unused set of credentials is non-trivial. Managing a ready-queue
    for each service is expensive in terms of space when there are many users.
    (See notes on sparsity in rateLimit.RateLimitEnforcer.) As an alternative,
    I rely on probability. (That is, until I think more and find a better
    solution.)

    The any_cred_queue maintains a circular buffer for each set of
    credentials. When a client asks for AnyCredentials, the queue pops
    credentials off the top. When it asks again, it pushes that initial item
    to the end of the queue, and pops the head again.

    This is far from perfect. For example, this could select a token that is
    currently rate-limited even though non-limited ones remain. But, it's
    cheap, and does well, probabilistically. Additionally, I believe this
    should perform better as the number of managed users rises. (I'll think
    more about it; then, pen an arXiv article.)
    """
    def __init__(self):
        self._credentials = {}
        self._any_cred_queue = []

    def is_managed(self, cred):
        return cred in self._credentials

    def add(self, cred):
        assert not self.is_managed(cred)

        client_req_processor = ClientRequestProcessor(cred)
        self._credentials[cred] = client_req_processor

        # Add a new set of credentials to the front of the queue.
        self._any_cred_queue.insert(0, cred)

        return client_req_processor

    def remove(self, cred):
        del self._credentials[cred]
        self._any_cred_queue.remove(cred)

    def __getitem__(self, cred):
        if cred is ANY_CREDENTIALS:
            credentials = self._any_cred_queue.pop(0)
            self._any_cred_queue.append(credentials)
            return self[credentials]
        else:
            managed = self._credentials.get(cred)

            if not managed:
                managed = self.add(cred)

            return managed

    def maintain(self, cull_interval=FIFTEEN_MINUTES):
        now = time.time()

        # Collect
        removal_keys = []
        for k, req_processor in self._credentials.items():
            req_processor.cleanup(now)  # Remove disused rate limits

            if req_processor.is_removable():
                removal_keys.append(k)

        # Remove
        for k in removal_keys:
            self.remove(k)


async def repeating(interval, f, *args, **kwargs):
    """
    Call f(**args, **kwargs) every interval seconds.

    :param interval: the interval in seconds
    :param f: a function
    :param args: positional arguments
    :param kwargs: keyword arguments
    """
    while True:
        f(*args, **kwargs)  # Do test for coro?
        await asyncio.sleep(interval)

# I don't want to require requests as a dependency. And, I really don't want
# to encourage the use of a blocking function, especially outside the system
# of rate limit management. But, when debugging, it's terribly useful to have
# a the following function.
try:
    from requests import request as _request
except ImportError:
    def _request(*args, **kwargs):
        raise ImportError("This function requires the `requests` package.")


def debug_blocking_request(app_cred, client_cred, twitter_req):
    """
    Fulfill a request using requests (for humans).

    WARNING: This method is for debugging purposes. It does no rate-limit
        enforcement.

    :param app_cred: the AppCredentials
    :param client_cred: the ClientCredentials
    :param twitter_req: The TwitterRequest

    :return: a requests.Response
    """
    headers = generate_req_headers(twitter_req, app_cred, client_cred)
    return _request(twitter_req.method,
                    twitter_req.url,
                    params=twitter_req.params,
                    headers=headers)



import unittest
from unittest.mock import MagicMock

from brittle_wit import TwitterRequest
from brittle_wit.executors import *
from brittle_wit.messages import TwitterError
from brittle_wit.oauth import AppCredentials, ClientCredentials
from test.helpers import *


class MockHTTPResp:
    def __init__(self, status, headers, body):
        self.status, self.headers, self.body = status, headers, body

    async def json(self):
        return self.body

    async def text(self):
        return self.body


class MockHTTPReq:
    def __init__(self, resp):
        self._resp = resp

    async def __aenter__(self):
        return self._resp

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


class TestTwitterReqToHTTPRequest(unittest.TestCase):
    def test_twitter_req_to_http_req(self):
        # This is a dumb test but it should help prevent dumb regressions.
        app_cred = AppCredentials("key", "secret")
        client_cred = ClientCredentials(1, "token-alice", "****")
        tr = TwitterRequest("POST", "http://url.com/", "service", "family")
        with aiohttp.ClientSession() as session:
            req = twitter_req_to_http_req(session, app_cred, client_cred, tr)
            self.assertEqual(type(req), aiohttp.client._RequestContextManager)


class TestExecuteRequest(unittest.TestCase):
    def test_good_resp_and_json_body(self):
        cred = ClientCredentials(1, "token-alice", "****")
        twitter_req = TwitterRequest("method", "url", "family", "service")
        rate_limit = RateLimit.from_ignorance()
        http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))

        coro = execute_req(cred, twitter_req, http_req, rate_limit, 1000)
        res = drive_coro_once(coro)
        self.assertIs(type(res), TwitterResponse)
        self.assertEqual(res.body['count'], 100)

    def test_good_resp_and_text_body(self):
        cred = ClientCredentials(1, "token-alice", "****")
        twitter_req = TwitterRequest("method", "url", "family", "service",
                                     parse_as='text')
        rate_limit = RateLimit.from_ignorance()
        http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))

        coro = execute_req(cred, twitter_req, http_req, rate_limit, 1000)
        res = drive_coro_once(coro)
        self.assertIs(type(res), TwitterResponse)
        self.assertEqual(res.body['count'], 100)

    def test_bad_resp(self):
        cred = ClientCredentials(1, "token-alice", "****")
        twitter_req = TwitterRequest("method", "url", "family", "service",
                                     parse_as='text')
        rate_limit = RateLimit.from_ignorance()
        http_req = MockHTTPReq(MockHTTPResp(403, {}, "Forbidden"))

        with self.assertRaises(TwitterError):
            coro = execute_req(cred, twitter_req, http_req, rate_limit, 1000)
            res = drive_coro_once(coro)


class TestClientRequestProcessor(unittest.TestCase):
    def setUp(self):
        self.app_cred = AppCredentials("key", "secret")
        self.credentials = [ClientCredentials(1, "token-alice", "****"),
                            ClientCredentials(2, "token-bob", "****"),
                            ClientCredentials(3, "token-carol", "****")]
        self.processor = ClientRequestProcessor(self.credentials)

    def test_init(self):
       self.assertTrue(self.processor.is_not_processing())

       tm = MagicMock()
       tm.all_lines_empty = lambda: False
       self.processor._ticket_master = tm

       self.assertFalse(self.processor.is_not_processing())

    def test_rate_limit_for(self):
        limit = self.processor._rate_limit_for("direct_messages")
        self.assertTrue(limit.is_pristine)
        self.assertIs(limit, self.processor._rate_limit_for("direct_messages"))

        self.assertIsNot(limit,
                         self.processor._rate_limit_for("account/settings"))

    def test_is_immediately_available_for(self):
        svc = "direct_message"
        self.assertTrue(self.processor.is_immediately_available_for(svc))

        limit = RateLimit.from_ignorance()
        limit._remaining = 0
        self.processor._rate_limits[svc] = limit
        self.assertFalse(self.processor.is_immediately_available_for(svc))

        tm = MagicMock()
        tm.is_line_empty = lambda s: False
        self.processor._ticket_master = tm
        self.assertFalse(self.processor.is_immediately_available_for(svc))

    def test_cleanup(self):
        self.assertTrue(self.processor.is_removable())
        self.assertFalse(self.processor._rate_limits)
        self.processor.cleanup()
        self.assertFalse(self.processor._rate_limits)

        now = time.time()
        self.processor._rate_limits['future'] = RateLimit(1, 1, now+100)
        self.processor._rate_limits['past'] = RateLimit(1, 1, now-100)
        self.assertEqual(set(self.processor._rate_limits), {'future', 'past'})
        self.processor.cleanup(now)
        self.assertFalse(self.processor.is_removable())
        self.assertEqual(set(self.processor._rate_limits), {'future'})
        self.assertFalse(self.processor.is_removable())


class TestManagedClientRequestProcessors(unittest.TestCase):
    def test_add_and_remove(self):
        alice = ClientCredentials(1, "alice", "secret")
        bob = ClientCredentials(2, "bob", "secret")
        processors = ManagedClientRequestProcessors()

        self.assertFalse(processors.is_managed(alice))
        self.assertFalse(processors.is_managed(bob))

        processors.add(alice)
        self.assertTrue(processors.is_managed(alice))
        self.assertFalse(processors.is_managed(bob))

        processors.remove(alice)
        self.assertFalse(processors.is_managed(alice))
        self.assertFalse(processors.is_managed(bob))

        processors.add(bob)
        self.assertFalse(processors.is_managed(alice))
        self.assertTrue(processors.is_managed(bob))

        with self.assertRaises(AssertionError):
            processors.add(bob)

    def test_get_item(self):
        alice = ClientCredentials(1, "alice", "secret")
        bob = ClientCredentials(2, "bob", "secret")
        processors = ManagedClientRequestProcessors()

        self.assertEqual(processors[alice]._client_credentials, alice)
        self.assertEqual(processors[bob]._client_credentials, bob)

        self.assertEqual({processors[ANY_CREDENTIALS]._client_credentials
                          for _ in range(10)},
                         {alice, bob})

    def test_maintain(self):
        alice = ClientCredentials(1, "alice", "secret")
        bob = ClientCredentials(2, "bob", "secret")
        processors = ManagedClientRequestProcessors()

        now = time.time()
        processors[alice]._rate_limits['future'] = RateLimit(1, 1, now+100)
        processors[bob]._rate_limits['passed'] = RateLimit(1, 1, now-100)

        self.assertTrue(processors.is_managed(alice))
        self.assertTrue(processors.is_managed(bob))

        processors.maintain()

        self.assertTrue(processors.is_managed(alice))
        self.assertFalse(processors.is_managed(bob))




import time
import unittest
from brittle_wit.rate_limit import RateLimit, TimingRateLimiter
from test.helpers import *


class TestRateLimit(unittest.TestCase):

    def test_from_ignorance(self):
        rate_limit = RateLimit.from_ignorance()
        self.assertFalse(rate_limit.is_exhausted)

    def test_from_initial_limit(self):
        rate_limit = RateLimit.from_initial_limit(10)
        self.assertEqual(rate_limit._remaining, 10)

    def test_from_response(self):
        resp = MockResp({'X-RATE-LIMIT-LIMIT': 100,
                         'X-RATE-LIMIT-REMAINING': 99,
                         'X-RATE-LIMIT-RESET': 9999})
        rate_limit = RateLimit.from_response(resp)
        expected = "RateLimit(limit=100, remaining=99, reset_time=21:46:39)"

        self.assertEqual(str(rate_limit), expected)
        self.assertEqual(rate_limit.__repr__(), expected)
        self.assertEqual(rate_limit.reset_time, 9999)

    def test_is_pristine(self):
        resp = MockResp({'X-RATE-LIMIT-LIMIT': 100,
                         'X-RATE-LIMIT-REMAINING': 100,
                         'X-RATE-LIMIT-RESET': 9999})
        rate_limit = RateLimit.from_response(resp)
        self.assertTrue(rate_limit.is_pristine)

        resp = MockResp({'X-RATE-LIMIT-REMAINING': 99})
        rate_limit.update(resp)
        self.assertFalse(rate_limit.is_pristine)

    def test_is_exhausted(self):
        rate_limit = RateLimit.from_ignorance()
        self.assertFalse(rate_limit.is_exhausted)

        resp = MockResp({'X-RATE-LIMIT-LIMIT': 100,
                         'X-RATE-LIMIT-REMAINING': 0,
                         'X-RATE-LIMIT-RESET': 9999})

        rate_limit.update(resp)
        self.assertTrue(rate_limit.is_exhausted)

    def test_bad_request_update(self):
        orig_resp = MockResp({'X-RATE-LIMIT-LIMIT': 100,
                              'X-RATE-LIMIT-REMAINING': 99,
                              'X-RATE-LIMIT-RESET': 9999})

        rate_limit = RateLimit.from_response(orig_resp)
        self.assertFalse(rate_limit.is_exhausted)
        rate_limit.update(MockResp({}))

    def test_mark_exhausted(self):
        orig_resp = MockResp({'X-RATE-LIMIT-LIMIT': 100,
                              'X-RATE-LIMIT-REMAINING': 99,
                              'X-RATE-LIMIT-RESET': 9999})

        rate_limit = RateLimit.from_ignorance()
        self.assertFalse(rate_limit.is_exhausted)
        rate_limit.mark_exhausted(orig_resp)
        self.assertTrue(rate_limit.is_exhausted)

    def test_comply_without_sleeping(self):
        rate_limit = RateLimit.from_ignorance()
        self.assertFalse(rate_limit.is_exhausted)
        drive_coro_once(rate_limit.comply())
        self.assertTrue(rate_limit.is_exhausted)

        rate_limit.update(MockResp({'X-RATE-LIMIT-LIMIT': 100,
                                    'X-RATE-LIMIT-REMAINING': 0,
                                    'X-RATE-LIMIT-RESET': time.time() + 0.1}))
        self.assertTrue(rate_limit.is_exhausted)
        drive_coro_once(rate_limit.comply())
        self.assertFalse(rate_limit.is_exhausted)


class TestTimingRateLimiter(unittest.TestCase):
    def test_timing_rate_limiter(self):
        limiter = TimingRateLimiter(60)
        self.assertTrue(limiter.can_proceed())
        self.assertFalse(limiter.can_proceed())

        limiter._start_time -= 61
        self.assertTrue(limiter.can_proceed())



import unittest
from brittle_wit.streaming import *
from brittle_wit.helpers import WrappingHandler


class TestEntryProcessor(unittest.TestCase):
    def test_entry_processor(self):
        ep = EntryProcessor()
        self.assertFalse(ep)
        self.assertEqual(ep.messages(), [])

        ep.process(b'"hello world"\r\n{"a": 20}')
        self.assertTrue(ep)
        self.assertEqual(ep.messages(), [b'"hello world"'])

        ep.process(b"\r\n42")
        self.assertEqual(ep.messages(), [b'{"a": 20}'])

        ep.process(b"\r\n")
        self.assertEqual(ep.messages(), [b"42"])
        self.assertEqual(ep.messages(), [])

        self.assertEqual(ep._buf, b"")
        ep.process(b"x")
        self.assertEqual(ep._buf, b"x")
        ep.process(b"kcd\r\n")
        self.assertEqual(ep.messages(), [b"xkcd"])
        self.assertEqual(ep._buf, b"")
        self.assertEqual(ep._buf, b"")

    def test_purge_mailbox(self):
        ep = EntryProcessor()
        ep.process(b"hello\r\nworld")
        self.assertTrue(ep)
        ep.purge_mailbox()
        self.assertFalse(ep)

    def test_purge_streaming(self):
        ep = EntryProcessor()
        ep.process(b"corrupt_buffer...")
        ep.purge_buffer()
        ep.process(b"good buffer\r\n")
        self.assertEqual(ep.take_one(), b"good buffer")

    def test_take_one(self):
        ep = EntryProcessor()
        ep.process(b"hello\r\nworld")
        self.assertEqual(ep.take_one(), b"hello")
        self.assertFalse(ep)

    def test_iteration(self):
        ep = EntryProcessor()
        ep.process(b"hello\r\nworld\r\n")
        self.assertEqual(list(ep), [b"hello", b"world"])


class TestParsers(unittest.TestCase):
    def test_noop_entry_parser(self):
        self.assertEqual(noop_entry_parser("hello"), "hello")

    def test_json_entry_parser(self):
        self.assertEqual(json_entry_parser(b"[10, 20, 30]"), [10, 20, 30])


class TestStreamingProcessor(unittest.TestCase):
    def test_streaming_processor(self):
        sp = StreamProcessor(None)
        self.assertFalse(sp._requires_json)

        json_msgs, byte_msgs = [], []
        json_handler = WrappingHandler(lambda m: json_msgs.append(m))
        byte_handler = WrappingHandler(lambda m: byte_msgs.append(m))

        sp.subscribe(byte_handler, False)
        self.assertFalse(sp._requires_json)
        sp.subscribe(json_handler, True)
        self.assertTrue(sp._requires_json)

        with self.assertRaises(RuntimeError):
            sp.subscribe(byte_handler, False)

        with self.assertRaises(RuntimeError):
            sp.subscribe(lambda: None, False)

        sp._send_to_all(b"[1,2,3]")

        self.assertEqual(json_msgs, [[1, 2, 3]])
        self.assertEqual(byte_msgs, [b"[1,2,3]"])

        sp.unsubscribe(json_handler)
        self.assertFalse(sp._requires_json)

        with self.assertRaises(RuntimeError):
            sp.unsubscribe(json_handler)



import asyncio


class Ticket:
    """
    A locking, synchronization context manager that enforces FIFO ordering.
    """

    __slots__ = '_ticket_master', '_line', '_antecedent', '_consequent'

    def __init__(self, ticket_master, line, antecedent_lock, consequent_lock):
        """
        :param ticket_master: the TicketMaster that originated this ticket
        :param line: the line this ticket is for
        :param antecedent_lock: the lock that must be acquired before the
            context body can execute
        :param consequent_lock: the lock that must be held until execution
            completes
        """
        self._ticket_master = ticket_master
        self._line = line
        self._antecedent = antecedent_lock
        self._consequent = consequent_lock

    @property
    def line(self):
        return self._line

    @property
    def consequent(self):
        return self._consequent

    async def acquire(self):
        # Immediately acquire the consequent lock. The next ticket
        # uses this as there antecedent. So, the next ticket cannot
        # progress until this ticket releases its consequent.
        await self._consequent.acquire()

        # Attempt to acquire the antecedent. This acts as a
        # barrier on execution. Continue once the antecedent
        # completes.
        async with self._antecedent:
            pass  # This is a barrier

    def release(self):
        # Release the consequent so that the next ticket can proceed.
        self._consequent.release()

        # Allow the TicketMaster to optionally cleanup a line.
        self._ticket_master.notify_finished(self)

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        self.release()




class TicketMaster:

    __slots__ = '_lines'

    def __init__(self):
        self._lines = {}  # line -> consequent

    def is_line_empty(self, line):
        """
        :return: True if the given line is empty
        """
        return line not in self._lines

    def all_lines_empty(self):
        """
        :return: True if the all the lines are empty
        """
        return not self._lines

    def notify_finished(self, ticket):
        """
        Notify this ticket master after ticket use.
        """
        if self._lines[ticket.line] is ticket.consequent:
            del self._lines[ticket.line]

    def take_ticket(self, line, loop=None):
        """
        Take a ticket on a specific line

        :param line: the requested line
        :param loop: the event loop

        :return: a Ticket
        """
        antecedent = self._lines.get(line, asyncio.Lock(loop=loop))
        consequent = asyncio.Lock(loop=loop)
        self._lines[line] = consequent

        return Ticket(self, line, antecedent, consequent)



import asyncio
import time

from datetime import datetime
from collections import defaultdict


FIFTEEN_MIN = 15 * 60 + 2 # Add two seconds.

class RateLimit:
    """
    Captures and enforces rate limiting.

    See: https://dev.twitter.com/rest/public/rate-limiting
    """
    CLIENT_LIMITS = defaultdict(lambda: FIFTEEN_MIN)
    APP_LIMITS = defaultdict(lambda: FIFTEEN_MIN)

    @staticmethod
    def from_ignorance():
        """
        Create a RateLimit object assuming one request allowed and remains.

        Given the way update works, this is a good initialization pattern.
        However, to make sure it is clear to the caller, ignorance is part of
        the name.
        """
        return RateLimit(1, 1, int(time.time() + FIFTEEN_MIN))  # 15 minutes from now

    @staticmethod
    def from_response(resp):
        """
        Create a RateLimit object from a twitter response.
        """
        return RateLimit.from_ignorance().update(resp)

    @staticmethod
    def from_initial_limit(initial_limit):
        """
        Create a RateLimit object from a limit with all requests remaining.
        """
        return RateLimit(initial_limit,
                         initial_limit,             # All remaining
                         int(time.time() + FIFTEEN_MIN))  # 15 minutes from now

    __slots__ = '_limit', '_remaining', '_reset_time'

    def __init__(self, limit, remaining, reset_time):
        self._limit = limit
        self._remaining = remaining
        self._reset_time = reset_time

    async def comply(self):
        """
        Ensure request availability before allowing caller to proceed.

        If there are zero requests remaining, comply uses sleep to wake up
        after the reset time. Call this method immediately preceding any
        request. Doing so allows for  maximum throughput (i.e. minimum wasted
        sleep time.)

        This method decrements the requests remaining prior to actually using a
        request. This gives strong guarantees on rate limit compliance. If the
        caller then uses the request then -- for some reason -- fails to call
        update, the RateLimit object remains correct. If the caller does call
        update, remaining requests comports to whatever is in the response
        object.
        """
        if self.is_exhausted:
            now = time.time()  # Is time.time() blocking?
            if now < self._reset_time:
                # Time sync issue so always add 5 seconds?
                await asyncio.sleep(self._reset_time - now + 5)
            self._remaining = self._limit

        self._remaining -= 1

    @property
    def is_exhausted(self):
        return self._remaining <= 0

    @property
    def is_pristine(self):
        return self._remaining == self._limit

    @property
    def reset_time(self):
        return self._reset_time

    def mark_exhausted(self, resp):
        """
        Ensure that the credentials are rate-limited following a bad response.

        From the Twitter Docs:

        > Lastly, there may be times in which the rate limit values that
        we return are inconsistent, or cases where no headers are returned at
        all.

        If there was an rate limit response, then there are definitely 0
        requests remaining. If the reset time is not present, then set it to
        15 minutes from now.

        :param resp:
        :return:
        """
        self._limit = int(resp.headers.get('X-RATE-LIMIT-LIMIT', self._limit))
        self._remaining = 0
        self._reset_time = int(resp.headers.get('X-RATE-LIMIT-RESET',
                                                time.time() + 60*15))

    def update(self, resp):
        """
        Update (overwrite) the limits given a Twitter response object.

        From the Twitter Docs:

        > Lastly, there may be times in which the rate limit values that
        we return are inconsistent, or cases where no headers are returned at
        all.

        To cope with this, I skip updates to missing header information.
        """
        self._limit = int(resp.headers.get('X-RATE-LIMIT-LIMIT', self._limit))
        self._remaining = int(resp.headers.get('X-RATE-LIMIT-REMAINING',
                                               self._remaining))
        self._reset_time = int(resp.headers.get('X-RATE-LIMIT-RESET',
                                                self._reset_time))

        return self

    def __str__(self):
        msg = "RateLimit(limit={}, remaining={}, reset_time={})"
        time_s = datetime.fromtimestamp(self._reset_time).strftime("%H:%M:%S")

        return msg.format(self._limit, self._remaining, time_s)

    def __repr__(self):
        return self.__str__()


class TimingRateLimiter:
    """
    Enforces a maximum number of operations per second.

    This class does not enforce Twitter's API rate limits. Instead, it's a soft
    limiter for social interactions. For example, if you are writing something
    ike a feed consolidator bot, you would not want to retweet more than an
    *average* of one status every hour. To do so, create a TimingRateLimiter
    with the minimum number of seconds between each operation.  Immediately
    prior to each possible operation, call bool(timing_rate_limiter). If it
    returns False, do not take the action. If it returns True, take the action.
    Implicitly, if it returns True, it assumes you did take the action.
    """
    def __init__(self, min_secs_per_op):
        """
        :param min_secs_per_op: the minimum average number of seconds which
            must pass between each operation
        """
        self._n = 0
        self._min_secs_per_op = min_secs_per_op
        self._start_time = time.time()

    @property
    def is_rate_limited(self):
        rate = (time.time() - self._start_time) / self._n
        return rate < self._min_secs_per_op

    def can_proceed(self):
        """
        :return: True if the operation is not limited.
        """
        if self._n == 0 or not self.is_rate_limited:
            self._n += 1
            return True
        else:
            return False



import json
import re
import textwrap
import warnings

from collections import OrderedDict, defaultdict
from inspect import Parameter, Signature
from itertools import tee, chain

from brittle_wit.rate_limit import RateLimit
from brittle_wit import TwitterRequest

SLUGS_RE = re.compile(r":[A-Za-z_0-9]+")
SLUG_POSTFIX_RE = re.compile(r"^.*?(:[A-Za-z_0-9]+)$")
SLUG_INFIX_RE = re.compile(r"""^(?:[A-Za-z]*/)?
                               ([A-Za-z0-9_]+)/
                               (\:[A-Za-z_0-9]+)/
                               (.*)$""",
                           re.X)

# Python cannot represent all Twitter API parameter names. They have invalid
# symbols. The following two dictionaries remap these names.
NAME_TO_PY_NAME = {"attribute:street_address": "attribute_street_address",
                   "media[]": "media_arr"}
PY_NAME_TO_NAME = {v: k for k, v in NAME_TO_PY_NAME.items()}


class NSSentinel:
    # XXX: There has to be an existing solution for this.
    def __repr__(self):
        return 'NOT_SPECIFIED'

NOT_SPECIFIED = NSSentinel()


def _generate_signature(api_def):
    """
    Generate a Python function signature from an api definition.

    This object enforces required and optional parameters. It also gives
    readable documentation and tab-completion. It's mostly meant as a
    guide though. It won't enforce strict correctness because the Twitter
    API has requirements like either A or B. In these cases, I mark both A
    and B as optional. If you don't specify one of them, Twitter will balk.
    But, it's reasonably obvious when this occurs, so it's not worth the
    added library complexity.
    """
    params = []

    for name, (optionality, _) in api_def['params'].items():
        name = NAME_TO_PY_NAME.get(name, name)

        if optionality == 'REQUIRED':
            p = Parameter(name, Parameter.POSITIONAL_OR_KEYWORD)
        else:
            p = Parameter(name, Parameter.KEYWORD_ONLY, default=NOT_SPECIFIED)

        params.append(p)

    return Signature(params)


def _generate_doc_str(api_def, line_width=70, indent_width=4):
    lines = textwrap.wrap(api_def['desc'], line_width) + [""]

    for k, (optionality, desc) in api_def['params'].items():
        lines.append(k + ":\n")

        body = "({}) {}".format(optionality, desc)
        for paragraph in body.splitlines():
            wrapped = textwrap.wrap(paragraph,
                                    line_width - indent_width,
                                    replace_whitespace=False)
            for line in wrapped:
                lines.append(" "*indent_width + line)
            lines.append("")

        lines.append('')

    return "\n".join(lines) + "\n"


def _update_params_ordering_required_first(api_def):
    """
    Updates the api_def IN PLACE, ensuring required params come first.

    The order of each parameter in `rest_api.json` is significant, matching
    the order in the API documentation. However, I can see a situation in
    which a required parameter comes after an optional one. That's fine for
    the documentation, but it's not okay for a Python signature.
    """
    t1, t2 = tee(api_def['params'].items())
    kv_pairs = chain((p for p in t1 if p[1][0] == 'REQUIRED'),
                     (p for p in t2 if p[1][0] != 'REQUIRED'))
    api_def['params'] = OrderedDict(kv_pairs)


def _bind_sig(signature, *args, **kwargs):
    return {PY_NAME_TO_NAME.get(k, k): v
            for k, v in signature.bind(*args, **kwargs).arguments.items()
            if v is not NOT_SPECIFIED}


def _slugged_pythonic_string(url):
    """
    Process a slugged URL

    :param url: the raw API endpoint
    :return: None if this is not a slugged url, otherwise a tuple of the
        python string for format-based interpolation and the slug parameter
        names
    """
    slugged_str = ""

    matches = list(SLUGS_RE.finditer(url))
    if not matches:
        return None

    last_idx = 0
    slug_params = []
    for m in matches:
        j, k = m.start(), m.end()
        slugged_str += url[last_idx:j]
        last_idx = k

        slug_param = m.group(0)[1:]
        slugged_str += "{" + slug_param + "}"
        slug_params.append(slug_param)

    return slugged_str, slug_params


def _generate_slugger(slugged_str, slug_params):
    """
    Create function that pops the slug_params and injects them into the URL.

    Note: The transform function takes the params dictionary, not the expanded
        **kwargs. The params dictionary is modified in place via pop!
    """
    def _slug_transform(params):
        slug_dict = {k: params.pop(k) for k in slug_params}
        return slugged_str.format(**slug_dict)
    return _slug_transform


def _url_to_func_name(service, family, method):
    # Special cases first.
    if service == "account/settings" and method == 'POST':
        return "update_settings"  # There is a 'GET' version

    m = SLUG_POSTFIX_RE.match(service)
    if m:
        postfix = m.group(1)
        service = service.replace(postfix, "by_" + postfix[1:])
    else:
        m = SLUG_INFIX_RE.match(service)
        if m:
            service = m.group(3) + "_by_" + m.group(1) + '_' + m.group(2)[1:]

    f_name = service.replace("/", "_")
    f_name = f_name.replace(family + '_', '')

    return f_name


def generate_api_request_builder_func(api_def):
    _update_params_ordering_required_first(api_def)

    f_name = _url_to_func_name(api_def['service'],
                               api_def['family'],
                               api_def['method'])

    signature = _generate_signature(api_def)

    def f(*args, **kwargs):
        # MAGIC WARNING
        binding = _bind_sig(signature, *args, **kwargs)
        return TwitterRequest(api_def['method'].upper(),
                              api_def['url'],
                              api_def['family'].upper(),
                              api_def['service'].upper(),
                              **binding)
    f.__name__ = f_name
    f.__doc__ = _generate_doc_str(api_def)
    f.__signature__ = signature

    return f


class TwitterAPI:
    def __init__(self, api_defs,
                 update_twitter_request_doc_urls=True,
                 update_rate_limit_defaults=True):

        if update_twitter_request_doc_urls:
            DOC_URLS = TwitterRequest.DOC_URLS

        if update_rate_limit_defaults:
            CLIENT_LIMITS = RateLimit.CLIENT_LIMITS
            APP_LIMITS = RateLimit.APP_LIMITS

        # Each API function belongs to a family.
        families = defaultdict(dict)

        # Build each API function, adding it to its family's dictionary.
        for api_def in api_defs:
            if 'deprecated' in api_def['service']:
                continue  # Nudge!

            f = generate_api_request_builder_func(api_def)
            if f.__name__ in families[api_def['family']]:
                warnings.warn("Conflicting on: " + api_def['reference_url'])
            if not f.__name__.isidentifier():
                fmt = "Bad Name: {} for {}"
                warnings.warn(fmt.format(f.__name__, api_def['reference_url']))
            families[api_def['family']][f.__name__] = f

            if update_twitter_request_doc_urls:
                DOC_URLS[api_def['url']] = api_def['reference_url']

            if update_rate_limit_defaults and api_def['rate_limited']:
                limits = api_def.get('limits')
                if limits:
                    if 'app' in limits:
                        APP_LIMITS[api_def['service']] = int(limits['app'])
                    if 'user' in limits:
                        CLIENT_LIMITS[api_def['service']] = int(limits['user'])

        # Make each family dictionary of api functions accessible via dot
        # accessor. That is, mock module-style access, without modules. There
        # are lots of ways to do this. But, I think creating a new type for
        # each family is the simplest. The str and repr encodings are easy to
        # grok from the REPL. And, the user is free to add end remove methods
        # to each family at will.
        for family_name, obj_dict in families.items():
            self.__dict__[family_name] = type(family_name, (object,), obj_dict)


def build_api(json_file):
    # The parameter ordering has meaning. It's copied from Twitter's API docs.
    decorder = json.JSONDecoder(object_pairs_hook=OrderedDict)
    with open(json_file) as fp:
        defs = decorder.decode(fp.read())
    return TwitterAPI(defs)



__version__ = "0.0.1"

import os


from brittle_wit.constants import LOGGER, FIFTEEN_MINUTES
from brittle_wit.messages import (TwitterRequest, Cursor, TwitterResponse,
                                  BrittleWitError)
from brittle_wit.oauth import (AppCredentials, ClientCredentials,
                               ANY_CREDENTIALS)
from brittle_wit.executors import (ClientRequestProcessor,
                                   ContextualizedProcessor,
                                   ManagedClientRequestProcessors)

from brittle_wit.helpers import parse_date
from brittle_wit.rest_api import build_api
from brittle_wit.streaming import (TwitterStream,
                                   StreamProcessor,
                                   StreamingHTTPPipe)


# By default, build each API endpoint into a function. Access the endpoints by
# dot-accessing the family. For example, `api.statuses.home_timeline()`. To
# disable API building -- mostly, if you are just issuing custom requests --
# set the API_BUILD environmental variable to 'SKIP'.
if os.environ.get('API_BUILD', '') != 'SKIP':
    LOGGER.debug("Building API from `api_reference.json`")
    api = build_api(os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                 "api_reference.json"))



from collections import namedtuple
from aiohttp.errors import ClientOSError, TimeoutError, ContentEncodingError
import asyncio

RETRYABLE_CODES = {500,  # INTERNAL SERVER ERROR
                   502,  # BAD GATEWAY
                   503,  # SERVICE UNAVAILABLE
                   504}  # GATEWAY TIMEOUT


RETRYABLE_EXCEPTIONS = {ContentEncodingError, ClientOSError, TimeoutError, 
                        asyncio.TimeoutError}


class TwitterRequest:
    """
    An Immutable (mostly) Twitter Request
    """

    # This is a static mapping from the request url to the Twitter API
    # documentation URL. `brittle_wit.twitter_request.build_api` populates it,
    # but there is no reason values cannot be overwritten.
    DOC_URLS = {}

    __slots__ = ('_method', '_url', '_family',
                 '_service', '_parse_as', '_params')

    def __init__(self, method, url, family, service,
                 *, parse_as='json', **params):
        self._method = method
        self._url = url
        self._family = family
        self._service = service.upper()
        self._parse_as = parse_as
        self._params = params

    @property
    def method(self):
        return self._method

    @property
    def url(self):
        return self._url

    @property
    def family(self):
        return self._family

    @property
    def service(self):
        return self._service

    @property
    def parse_as(self):
        return self._parse_as

    @parse_as.setter
    def parse_as(self, parse_format):
        self._parse_as = parse_format

    @property
    def params(self):
        return self._params

    def __str__(self):
        return "TwitterRequest(url={})".format(self._url)

    def clone_and_merge(self, updated_params):
        return TwitterRequest(self._method, self._url, self._family,
                              self._service,
                              parse_as=self._parse_as,
                              **{**self._params, **updated_params})


class Cursor:
    def __init__(self, twitter_req):
        self._req = twitter_req
        self._cursor = -1
        self._update_called = True

    def update(self, resp):
        next_cursor = resp.body.get('next_cursor')

        if next_cursor is None or next_cursor == 0:
            self._req = None
        else:
            self._cursor = next_cursor
            self._req = self._req.clone_and_merge({'cursor': next_cursor})

        self._update_called = True

    def __iter__(self):
        return self

    def __next__(self):
        if not self._update_called:
            raise RuntimeError("Must call update() between __next__s")

        if self._req is None:
            raise StopIteration

        self._update_called = False
        return self._req


TwitterResponse = namedtuple('TwitterResponse', 'req resp body')


class BrittleWitError(Exception):
    @property
    def is_retryable(self):
        return False


class TwitterError(BrittleWitError):
    def __init__(self, credentials, twitter_req, http_resp, message):
        self._credentials = credentials
        self._twitter_req = twitter_req
        self._http_resp = http_resp
        self._status_code = http_resp.status
        self._message = message

    @property
    def status_code(self):
        return self._status_code

    @property
    def twitter_req(self):
        return self._twitter_req

    @property
    def is_retryable(self):
        return self._status_code in RETRYABLE_CODES

    @property
    def message(self):
        return self._message

    @property
    def credentials(self):
        return self._credentials

    def __str__(self):
        return "TwitterError(code={}, msg={})".format(self._status_code,
                                                      self._message)

    def __repr__(self):
        return self.__str__()


class WrappedException(BrittleWitError):
    def __init__(self, underlying_exception):
        self._underlying_exception = underlying_exception

    @property
    def underlying_exception(self):
        return self._underlying_exception

    @property
    def is_retryable(self):
        return type(self._underlying_exception) in RETRYABLE_EXCEPTIONS

    def __str__(self):
        fmt = "WrappedException({})"
        return fmt.format(repr(self._underlying_exception))

    def __repr__(self):
        return self.__str__()



from distutils.core import setup

setup(
    name="brittle_wit",
    version="0.0.2",
    author="John Bjorn Nelson",
    author_email="jbn@pathdependent.com",
    packages=["brittle_wit",],
    keywords=["twitter"],
    license="MIT",
    description="An async Twitter library for Python.",
    long_description=open("README.md").read(),
    url="https://github.com/jbn/BrittleWit",
    package_data={"brittle_wit": ["api_reference.json"]},
    include_package_data=True,
)



import unittest
import brittle_wit as bw
from brittle_wit import rest_api
from brittle_wit.rest_api import _slugged_pythonic_string, _generate_slugger
from collections import OrderedDict


class TestRestAPI(unittest.TestCase):

    def test_geo_search(self):
        # Note attribute_street_address -> attribute:street_address
        req = bw.api.geo.search(attribute_street_address="sullivan street")
        self.assertEqual(req.method, "GET")
        self.assertEqual(req.family, "GEO")
        self.assertEqual(req.url,
                         "https://api.twitter.com/1.1/geo/search.json")
        self.assertIn("attribute:street_address", req.params)
        self.assertEqual(req.params["attribute:street_address"],
                         "sullivan street")

    def test_update_params_ordering_required_first(self):
        api_def = dict(params=OrderedDict([("a", ('REQUIRED', "desc")),
                                           ("b", ('OPTIONAL', "desc")),
                                           ("c", ('OPTIONAL', "desc")),
                                           ("d", ('REQUIRED', "desc"))]))

        rest_api._update_params_ordering_required_first(api_def)
        keys = list(api_def['params'].keys())
        self.assertEqual(set(keys[:2]), {"a", "d"})
        self.assertEqual(set(keys[2:]), {"b", "c"})

    def test_slugged_pythonic_string(self):
        examples = [("https://api.twitter.com/1.1/geo/id/:place_id.json",
                     ('https://api.twitter.com/1.1/geo/id/{place_id}',
                      ['place_id'])),
                    ("https://api.twitter.com/1.1/:user_id/id/:place_id.json",
                     ('https://api.twitter.com/1.1/{user_id}/id/{place_id}',
                      ['user_id', 'place_id'])),
                    ("https://api.twitter.com/1.1/geo/id/home",
                     None)]

        for url, expected in examples:
            self.assertEqual(_slugged_pythonic_string(url),
                             expected)

    def test_generate_slugger(self):
        f = _generate_slugger('https://api.twitter.com/1.1/geo/id/{pid}',
                              ['pid'])
        params = dict(pid=100)
        self.assertIn('pid', params)
        self.assertEqual(f(params), "https://api.twitter.com/1.1/geo/id/100")
        self.assertNotIn('pid', params)

        f = _generate_slugger('https://api.twitter.com/1.1/{gid}/id/{pid}',
                              ['gid', 'pid'])
        params = dict(pid=100, gid=200)
        self.assertIn('pid', params)
        self.assertIn('gid', params)
        self.assertEqual(f(params), "https://api.twitter.com/1.1/200/id/100")
        self.assertNotIn('pid', params)
        self.assertNotIn('gid', params)



import asyncio
import json

import aiohttp
from aiohttp import web

from brittle_wit.constants import LOGGER
from brittle_wit.executors import twitter_req_to_http_req
from brittle_wit.messages import TwitterError, BrittleWitError


class EntryProcessor:
    """
    Splits a stream of bytes into CRLF-separated messages.
    """

    def __init__(self):
        self._buf = b""  # Use a more efficient buffer
        self._mailbox = []

    def process(self, chunk):
        """
        Translate received bytes into messages in a mailbox.

        :param chunk: a chunk of bytes received from the HTTP connection
        """
        self._buf += chunk

        # Streams are `\r\n`-separated JSON messages.
        raw_lines = self._buf.split(b"\r\n")

        # If only one element in the split, then there wasn't a CRLF.
        if len(raw_lines) > 1:
            # The last element may be a b'', which is perfectly fine.
            self._buf = raw_lines[-1]

            # Blank lines are keep-alive messages.
            self._mailbox.extend(l for l in raw_lines[:-1] if l.strip())

    def messages(self):
        """
        Return the messages then reset the mailbox

        :return: the enqueued messages
        """
        msgs = self._mailbox
        self._mailbox = []
        return msgs

    def __bool__(self):
        """
        :return: True if there are any messages available
        """
        return bool(self._mailbox)

    def take_one(self):
        """
        :return: the first message removed from the mailbox
        """
        return self._mailbox.pop(0)

    def purge_buffer(self):
        """
        Reset the internal buffer.

        This is useful when restarting a stream. When the old stream
        terminates, the buffer is incomplete. Appending to it will produce a
        corrupt entry.
        """
        self._buf = b''

    def purge_mailbox(self):
        """
        Resets the mailbox so there are no messages
        """
        self._mailbox = []

    def __iter__(self):
        """
        :return: an iterator which pops messages of the mailbox until fully
            consumed
        """
        return self._entry_consumer()

    def _entry_consumer(self):
        while self._mailbox:
            yield self.take_one()


def noop_entry_parser(entry):
    """
    :return: the entry string without further processing.
    """
    return entry


def json_entry_parser(entry):
    """
    :return: the entry parsed as a JSON object
    """
    return json.loads(entry.decode('ascii'))


class TwitterStream:
    def __init__(self, session, app_cred, client_cred, twitter_req,
                 parser=noop_entry_parser, chunk_size=4096):
        """
        Initialize a Twitter stream.

        This connects to Twitter's servers but doesn't start reading until a
        call to __iter__.

        :param session: the aiohttp.ClientSession object
        :param app_cred: the AppCredentials
        :param client_cred: the ClientCredentials
        :param twitter_req: the TwitterRequest to initiate streaming
        :param parser: a parser which takes a byte-entry and translates it
            into a message.
        :param chunk_size: the number of bytes to asynchronously read from
            Twitter at a time
        """
        self._session = session
        self._app_cred = app_cred
        self._client_cred = client_cred
        self._twitter_req = twitter_req
        self._parser = parser
        self._chunk_size = chunk_size

        self._entry_processor = EntryProcessor()
        self._http_req = None

        self._connect()

    def _connect(self):
        """
        Connect to Twitter's servers.
        """
        # There may only be one set of credentials connected to a
        # streaming endpoint at a time. Preempting an existing
        # connection is error-prone. Complain loudly when a connection
        # already exists.
        if self.is_open:
            raise RuntimeError("Already connected. Call close() first!")

        self._http_req = twitter_req_to_http_req(self._session,
                                                 self._app_cred,
                                                 self._client_cred,
                                                 self._twitter_req)

    @property
    def is_open(self):
        """
        :return: True if their is an active connection to Twitter's servers.
        """
        return self._http_req is not None and not self._resp.connection.closed

    def reconnect(self, force_close_if_open=False, clear_prior_messages=False):
        """
        Reconnect to Twitter's servers.

        :param force_close_if_open: if True, close the prior connection, if it
            exists
        :param clear_prior_messages: if True, purges all (correct) messages
            waiting for consumption on the internal EntryProcessor from the
            prior connection
        """
        # Explicitly acknowledge that forced closure is okay.
        if force_close_if_open and self.is_open:
            self.disconnect()

        # The initiating request does not change between connection resets,
        # so old messages are consumable.
        if clear_prior_messages:
            self._entry_processor.purge_mailbox()

        # The buffer underling the entry processor should always be in
        # an initially pristine state. A broken connection virtually
        # ensures corruption.
        self._entry_processor.purge_buffer()

        self._connect()

    def disconnect(self):
        """
        Disconnect the stream from Twitter's servers.
        """
        if self._http_req is not None:
            self._http_req.close()
            self._http_req = None

    async def __aiter__(self):
        """
        Commence streaming, returning this object as an iterator.
        """
        # I hate the next line. Is there a decontextualize pattern?
        self._resp = await self._http_req.__aenter__()
        if self._resp.status != 200:
            raise TwitterError(self._client_cred,
                               self._twitter_req,
                               self._resp,
                               await self._resp.text())

        return self

    async def __anext__(self):
        # If there are no messages ready, read from the stream.
        while not self._entry_processor:
            chunk = await self._resp.content.read(self._chunk_size)
            if not chunk:
                LOGGER.error("Read zero bytes on a stream.")
                break
            self._entry_processor.process(chunk)

        # If there are no messages ready, the stream closed!
        if not self._entry_processor:
            LOGGER.error("Stream Closing (non-explicit)")
            raise StopAsyncIteration

        # Take one message.
        return self._parser(self._entry_processor.take_one())


class StreamProcessor:
    def __init__(self, twitter_stream):
        self._twitter_stream = twitter_stream
        self._subscribers = {}
        self._requires_json = False

    def subscribe(self, handler, as_json=False):
        handler_id = id(handler)

        if handler_id in self._subscribers:
            raise RuntimeError("Already subscribed!")

        if not hasattr(handler, "send"):
            raise RuntimeError("Handler must implement send(msg)")

        if as_json:
            self._requires_json = True

        self._subscribers[handler_id] = (handler, as_json)
        return handler_id

    def unsubscribe(self, handler):
        handler_id = id(handler)

        if handler_id not in self._subscribers:
            msg = "Either never subscribed or already unsubscribed!"
            raise RuntimeError(msg)

        required_json = self._subscribers[handler_id][1]

        del self._subscribers[handler_id]

        if required_json and not any(p[1] for p in self._subscribers.values()):
            self._requires_json = False

    def _send_to_all(self, message):
        # JSON parsing is expensive. If the application only pipes the
        # Twitter Stream to HTTP clients, it's dead time. The pattern
        # would be: BYTES -> JSON -> BYTES
        if self._requires_json:
            json_message = json_entry_parser(message)

        for subscriber, as_json in self._subscribers.values():
            subscriber.send(json_message if as_json else message)

    async def run(self):
        # Retry Strategy: https://dev.twitter.com/streaming/overview/connecting
        # TODO: More compliant!
        failures, failed = 0, False
        while True:
            try:
                if failed:
                    failures += 1
                    self._twitter_stream.reconnect(force_close_if_open=True)
                async for message in self._twitter_stream:
                    self._send_to_all(message)
                    failed = False  # TODO: FIX: This is such a waste!
            except TwitterError as e:
                # The docs only say 420. But, I suspect 429 codes, too.
                if e.status_code == 420 or e.status_code == 429:
                    sleep_time = 60 * 2**failures
                    msg = "Stream 420'd waiting {} seconds (failure {})"
                    LOGGER.error(msg.format(sleep_time, failures))
                    await asyncio.sleep(sleep_time)
                else:
                    sleep_time = min(5 * 2**failures, 320)
                    msg = "Stream {}'d waiting {} seconds (failure {})"
                    LOGGER.error(msg.format(e.status_code, sleep_time,
                                            failures))
                    print(e._http_resp)  # FIX: XXX: HACK: FUCK

                failed = True
                await asyncio.sleep(sleep_time)
            except BrittleWitError as e:
                if e.is_retryable:  # TODO FIX as is network error.
                    sleep_time = min(0.25 * failures, 16)
                    msg = "Stream {}'d waiting {} seconds (failure {})"
                    LOGGER.error(msg.format(repr(e), sleep_time, failures))
                    failed = True
                    await asyncio.sleep(sleep_time)
            except KeyboardInterrupt:
                raise
            except Exception as e:
                print(e)
                sleep_time = min(0.25 * failures, 16)
                msg = "Stream {}'d waiting {} seconds (failure {})"
                LOGGER.error(msg.format(repr(e), sleep_time, failures))
                failed = True
                await asyncio.sleep(sleep_time)


class StreamingHTTPPipe:
    DEFAULT_HEADERS = {'Content-Type': 'application/json'}

    def __init__(self, **resp_headers):
        # The _clients dict maps a unique identifier -- via id(obj) -- to a
        # pair of (StreamResponse, Event).
        self._clients = {}

        self._headers = {**StreamingHTTPPipe.DEFAULT_HEADERS, **resp_headers}

    async def handle(self, req):
        caller_id = id(req)
        resp = web.StreamResponse(status=200, headers=self._headers)

        try:
            # Start the FSM.
            await resp.prepare(req)

            # The send method signals handle when it processing finishes.
            finished_flag = asyncio.Event()
            self._clients[caller_id] = (resp, finished_flag)
            await finished_flag.wait()

        except asyncio.CancelledError:
            # Generally, this means the client disconnected.
            if caller_id in self._clients:
                del self._clients[caller_id]

        return resp

    def send(self, message):
        removal_set = []

        for caller_id, (resp, finished_flag) in self._clients.items():
            try:
                resp.write(message)  # I don't think draining is nessessary.
            except Exception as e:
                # This is a sloppy catch all for now. I'm not sure if
                # asyncio guarantees a cancel to handle prior to a
                # possible write. But, I do know it would be quite
                # wrong for a write to one client to cause failures
                # in other clients.
                print("Send", repr(e))
                finished_flag.set()
                removal_set.append(caller_id)
                raise e

        for k in removal_set:
            del self._clients[k]


async def save_raw_stream(session, app_cred, client_cred, twitter_req,
                          output_path, chunk_size=4096, timeout=30):
    """
    Save response headers and timeout seconds of response for a streaming req.

    I use this function to save some stream output so I can write text fixtures
    against the streaming API.

    :param session: the aiohttp ClientSession
    :param app_cred: an AppCredentials object
    :param client_cred: a ClientCredentials object
    :param twitter_req: the TwitterRequest for the streaming endpoint
    :param output_path: the output path. Technically, this is a prefix as the
        function writes two output files: prefix + '.headers.json' and
        prefix + '.content.raw'
    :param chunk_size: the number of bytes for read chunking
    :param timeout: the number of seconds to pipe response to file output
        before exiting.
    """
    header_path = output_path + ".headers.json"
    content_path = output_path + ".content.raw"

    http_req = twitter_req_to_http_req(session, app_cred,
                                       client_cred, twitter_req)

    # This uses blocking file operations. It's possible that a slow
    # disk could generate a stale warning. But, that's fine for the
    # use case of this function: debugging. A stale warning is useful
    # information to have in a saved feed as a fixture.
    with open(header_path, "w") as hp, open(content_path, "wb") as fp:
        try:
            with aiohttp.Timeout(timeout):
                async with http_req as resp:
                    # Save response headers.
                    json.dump({k: v for k, v in resp.headers.items()},
                              hp, indent=4)

                    while True:
                        chunk = await resp.content.read(chunk_size)
                        if not chunk:
                            break
                        fp.write(chunk)
        except asyncio.TimeoutError:
            pass



import unittest

from brittle_wit.helpers import *


class TestHelpers(unittest.TestCase):

    def test_parse_date(self):
        s = "Tue Mar 29 15:40:03 +0000 2016"
        self.assertEqual(parse_date(s).timestamp(), 1459266003.0)

    def test_wrapping_handler(self):
        messages = []

        handler = WrappingHandler(lambda msg: messages.append(msg))
        handler.send("hello.")
        handler.send("it's me.")

        self.assertEqual(" ".join(messages), "hello. it's me.")
